import time

import requests

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        SESSION_URL = "https://my.kirklees.gov.uk/authapi/isauthenticated?uri=https%253A%252F%252Fmy.kirklees.gov.uk%252Fservice%252FBins_and_recycling___Manage_your_bins&hostname=my.kirklees.gov.uk&withCredentials=true"

        API_URL = "https://my.kirklees.gov.uk/apibroker/runLookup"

        data = {
            "formValues": {"Search": {"validatedUPRN": {"value": user_uprn}}},
        }
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": "Mozilla/5.0",
            "X-Requested-With": "XMLHttpRequest",
            "Referer": "https://my.kirklees.gov.uk/fillform/?iframe_id=fillform-frame-1&db_id=",
        }
        s = requests.session()
        r = s.get(SESSION_URL)
        r.raise_for_status()
        session_data = r.json()
        sid = session_data["auth-session"]
        params = {
            "id": "65e08e60b299d",
            "repeat_against": "",
            "noRetry": "false",
            "getOnlyTokens": "undefined",
            "log_id": "",
            "app_name": "AF-Renderer::Self",
            # unix_timestamp
            "_": str(int(time.time() * 1000)),
            "sid": sid,
        }

        r = s.post(API_URL, json=data, headers=headers, params=params)
        r.raise_for_status()

        data = r.json()
        rows_data = data["integration"]["transformed"]["rows_data"]
        if not isinstance(rows_data, dict):
            raise ValueError("Invalid data returned from API")

        for bin_id, bin_info in rows_data.items():
            label = bin_info.get("label", "Unknown")
            next_collection_date = bin_info.get("NextCollectionDate", "Unknown")
            # Convert the date string into a readable format
            try:
                formatted_date = datetime.strptime(
                    next_collection_date, "%Y-%m-%dT%H:%M:%S"
                ).strftime(date_format)
            except ValueError:
                formatted_date = "Unknown"

            dict_data = {"type": label, "collectionDate": formatted_date}
            bindata["bins"].append(dict_data)

        return bindata



import time

import requests
from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        URI = "https://lbhapiprod.azure-api.net"
        endpoint = f"{URI}/whitespace/GetCollectionByUprnAndDate"
        subscription_key = "2ea6a75f9ea34bb58d299a0c9f84e72e"

        # Get today's date in 'YYYY-MM-DD' format
        collection_date = datetime.now().strftime("%Y-%m-%d")

        # Define the request headers
        headers = {
            "Content-Type": "application/json",
            "Ocp-Apim-Subscription-Key": subscription_key,
        }

        # Define the request body
        data = {
            "getCollectionByUprnAndDate": {
                "getCollectionByUprnAndDateInput": {
                    "uprn": user_uprn,
                    "nextCollectionFromDate": collection_date,
                }
            }
        }
        # Make the POST request
        response = requests.post(endpoint, headers=headers, data=json.dumps(data))
        response.raise_for_status()  # Raise an exception for HTTP errors

        # Parse the JSON response
        response_data = response.json()

        collections = (
            response_data.get("getCollectionByUprnAndDateResponse", {})
            .get("getCollectionByUprnAndDateResult", {})
            .get("Collections", [])
        )

        for collection in collections:
            bin_type = collection["service"]
            collection_date = collection["date"]

            dict_data = {
                "type": bin_type,
                "collectionDate": datetime.strptime(
                    collection_date,
                    "%d/%m/%Y %H:%M:%S",
                ).strftime(date_format),
            }
            bindata["bins"].append(dict_data)
        bindata["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), date_format)
        )

        return bindata



import requests
from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        URI = f"https://www.west-dunbarton.gov.uk/recycling-and-waste/bin-collection-day/?uprn={user_uprn}"

        # Make the GET request
        response = requests.get(URI)

        soup = BeautifulSoup(response.content, "html.parser")

        # For each next-date class get the text within the date-string class
        schedule_details = soup.findAll("div", {"class": "round-info"})

        for item in schedule_details:
            schedule_date = item.find("span", {"class": "date-string"}).text.strip()
            schedule_type = item.find("div", {"class": "round-name"}).text.strip()
            # Format is 22 March 2023 - convert to date
            collection_date = datetime.strptime(schedule_date, "%d %B %Y").date()

            # If the type contains "Blue bin or bag" or "Blue" then set the type to "BLUE"
            if "bag" in schedule_type.lower() or "blue" in schedule_type.lower():
                dict_data = {
                    "type": "Blue",
                    "collectionDate": collection_date.strftime(date_format),
                }
                bindata["bins"].append(dict_data)

            # If the type contains "caddy" or "brown" then set the type to "BROWN"
            if "caddy" in schedule_type.lower() or "brown" in schedule_type.lower():
                dict_data = {
                    "type": "Brown",
                    "collectionDate": collection_date.strftime(date_format),
                }
                bindata["bins"].append(dict_data)

            # If the type contains "Non-Recyclable" then set the type to "BLACK", compare in lowecase
            if "non-recyclable" in schedule_type.lower():
                dict_data = {
                    "type": "Black",
                    "collectionDate": collection_date.strftime(date_format),
                }
                bindata["bins"].append(dict_data)

        bindata["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), "%d/%m/%Y")
        )

        return bindata



from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_date(self, date_str):
        months = {
            "January": "01",
            "February": "02",
            "March": "03",
            "April": "04",
            "May": "05",
            "June": "06",
            "July": "07",
            "August": "08",
            "September": "09",
            "October": "10",
            "November": "11",
            "December": "12",
        }
        day, date, month_abbr, year = date_str.split()
        month = months[month_abbr]
        return f"{date}/{month}/{year}"

    def add_bin_types_to_collection(
        self, bin_data: {"bins": []}, collection_date: str, collectionType: str
    ):
        if "Grey Bin" in collectionType:
            bin_data["bins"].append(
                {
                    "type": "Grey Bin",
                    "collectionDate": self.parse_date(collection_date),
                }
            )
        if "Green Bin" in collectionType:
            bin_data["bins"].append(
                {
                    "type": "Green Bin",
                    "collectionDate": self.parse_date(collection_date),
                }
            )

        if "Blue Bin" in collectionType:
            bin_data["bins"].append(
                {
                    "type": "Blue Bin",
                    "collectionDate": self.parse_date(collection_date),
                }
            )

    def parse_data(self, page: str, **kwargs) -> dict:
        # Make a BS4 object
        soup = BeautifulSoup(page.text, features="html.parser")
        soup.prettify()

        # Initialize the bin data structure
        bin_data = {"bins": []}

        collectionDatesSection = soup.find("div", id="showCollectionDates")

        # Find next date
        collection_date = collectionDatesSection.find(
            "p", class_="collection-date"
        ).getText()

        # convert to date
        collection_type = collectionDatesSection.find(
            "p", class_="collection-type"
        ).getText()

        self.add_bin_types_to_collection(bin_data, collection_date, collection_type)

        # Find the table with collection dates
        table = collectionDatesSection.find("table", class_="leisure-table")

        # Extract the rows containing the bin collection information
        rows = table.find_all("tr")

        # Loop through the rows and extract bin data
        for row in rows:
            cells = row.find_all("td")
            if len(cells) == 2:
                collection_date = cells[1].get_text(strip=True)
                collection_type = cells[0].get_text(strip=True)

                self.add_bin_types_to_collection(
                    bin_data, collection_date, collection_type
                )

        return bin_data



from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        collection_day = kwargs.get("paon")
        round = kwargs.get("postcode")

        bindata = {"bins": []}

        days_of_week = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]

        round_week = ["Round A", "Round B"]

        offset_days = days_of_week.index(collection_day)
        round_collection = round_week.index(round)

        if round_collection == 0:
            bluebrownstartDate = datetime(2024, 10, 21)
            greengreystartDate = datetime(2024, 10, 28)
        else:
            greengreystartDate = datetime(2024, 10, 21)
            bluebrownstartDate = datetime(2024, 10, 28)

        greengrey_dates = get_dates_every_x_days(greengreystartDate, 14, 28)
        bluebrown_dates = get_dates_every_x_days(bluebrownstartDate, 14, 28)
        food_dates = get_dates_every_x_days(greengreystartDate, 7, 56)

        for greengrey_date in greengrey_dates:

            collection_date = (
                datetime.strptime(greengrey_date, "%d/%m/%Y")
                + timedelta(days=offset_days)
            ).strftime("%d/%m/%Y")

            dict_data = {
                "type": "Green/Grey Bin",
                "collectionDate": collection_date,
            }
            bindata["bins"].append(dict_data)

        for bluebrown_date in bluebrown_dates:

            collection_date = (
                datetime.strptime(bluebrown_date, "%d/%m/%Y")
                + timedelta(days=offset_days)
            ).strftime("%d/%m/%Y")

            dict_data = {
                "type": "Blue Bin",
                "collectionDate": collection_date,
            }
            bindata["bins"].append(dict_data)
            dict_data = {
                "type": "Brown Bin",
                "collectionDate": collection_date,
            }
            bindata["bins"].append(dict_data)

        for food_date in food_dates:

            collection_date = (
                datetime.strptime(food_date, "%d/%m/%Y") + timedelta(days=offset_days)
            ).strftime("%d/%m/%Y")

            dict_data = {
                "type": "Food Bin",
                "collectionDate": collection_date,
            }
            bindata["bins"].append(dict_data)

        bindata["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), "%d/%m/%Y")
        )

        return bindata



import time

import requests

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        data = {"uprn": user_uprn, "usrn": "1"}

        URI = (
            "https://zbr7r13ke2.execute-api.eu-west-2.amazonaws.com/Beta/get-bin-dates"
        )

        # Make the GET request
        response = requests.post(URI, json=data)
        response.raise_for_status()

        # Parse the JSON response
        bin_collection = json.loads(response.json()["dates"])
        collections = {
            "General": bin_collection["blackBinDay"],
            "Recycling": bin_collection["recyclingBinDay"],
            "Food": bin_collection["foodBinDay"],
            "Garden": bin_collection["gardenBinDay"],
        }
        # Loop through each collection in bin_collection
        for collection in collections:
            print(collection)

            if len(collections[collection]) <= 0:
                continue
            for date in collections[collection]:
                date = (
                    datetime.strptime(date, "%Y-%m-%dT%H:%M:%S").strftime("%d/%m/%Y"),
                )
                dict_data = {"type": collection, "collectionDate": date[0]}
                bindata["bins"].append(dict_data)

        return bindata



from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        URI = f"https://www.teignbridge.gov.uk/repositories/hidden-pages/bin-finder?uprn={user_uprn}"

        driver = create_webdriver(web_driver, headless, None, __name__)
        driver.get(URI)

        soup = BeautifulSoup(driver.page_source, features="html.parser")

        collection_dates = soup.find_all(
            "h3"
        )  # Assuming bin types are inside <h3> tags
        bin_type_headers = soup.find_all(
            "div", {"class": "binInfoContainer"}
        )  # Assuming collection dates are inside <p> tags

        # Iterate over the results and extract bin type and collection dates
        for i, date in enumerate(collection_dates):
            collection_date = date.get_text(strip=True)

            bin_types = bin_type_headers[i].find_all("div")
            for bin_type in bin_types:
                dict_data = {
                    "type": bin_type.text.strip(),
                    "collectionDate": datetime.strptime(
                        collection_date,
                        "%d %B %Y%A",
                    ).strftime("%d/%m/%Y"),
                }
                bindata["bins"].append(dict_data)

        bindata["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), "%d/%m/%Y")
        )

        return bindata



from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass
from selenium import webdriver
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from datetime import datetime
from uk_bin_collection.uk_bin_collection.common import date_format


class CouncilClass(AbstractGetBinDataClass):
    def parse_data(self, page: str, **kwargs) -> dict:
        postcode = kwargs.get("postcode", "")
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")

        options = webdriver.ChromeOptions()
        if headless:
            options.add_argument("--headless")
        driver = create_webdriver(web_driver, headless)

        try:
            driver.get(
                f"https://eppingforestdc.maps.arcgis.com/apps/instant/lookup/index.html?appid=bfca32b46e2a47cd9c0a84f2d8cdde17&find={postcode}"
            )
            wait = WebDriverWait(driver, 10)
            WebDriverWait(driver, 10).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, ".esri-feature-content")
                )
            )
            html_content = driver.page_source

            soup = BeautifulSoup(html_content, "html.parser")
            bin_info_divs = soup.select(".esri-feature-content p")
            data = {"bins": []}
            for div in bin_info_divs:
                if "collection day is" in div.text:
                    bin_type, date_str = div.text.split(" collection day is ")
                    bin_dates = datetime.strptime(
                        date_str.strip(), "%d/%m/%Y"
                    ).strftime(date_format)
                    data["bins"].append(
                        {"type": bin_type.strip(), "collectionDate": bin_dates}
                    )

            return data
        finally:
            driver.quit()



from datetime import datetime

import requests
from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Get and check UPRN
        user_uprn = kwargs.get("uprn")
        user_postcode = kwargs.get("postcode")
        check_uprn(user_uprn)
        check_postcode(user_postcode)
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")
        bindata = {"bins": []}

        API_URL = "https://secure.ashford.gov.uk/waste/collectiondaylookup/"

        # Create Selenium webdriver
        driver = create_webdriver(web_driver, headless, None, __name__)
        driver.get(API_URL)

        # Wait for the postcode field to appear then populate it
        inputElement_postcode = WebDriverWait(driver, 30).until(
            EC.presence_of_element_located(
                (By.ID, "ContentPlaceHolder1_CollectionDayLookup2_TextBox_PostCode")
            )
        )
        inputElement_postcode.send_keys(user_postcode)

        # Click search button
        findAddress = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located(
                (
                    By.ID,
                    "ContentPlaceHolder1_CollectionDayLookup2_Button_PostCodeSearch",
                )
            )
        )
        findAddress.click()

        # Wait for the 'Select your property' dropdown to appear and select the first result
        dropdown = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.ID,
                    "ContentPlaceHolder1_CollectionDayLookup2_DropDownList_Addresses",
                )
            )
        )

        # Create a 'Select' for it, then select the first address in the list
        # (Index 0 is "Make a selection from the list")
        dropdownSelect = Select(dropdown)
        dropdownSelect.select_by_value(str(user_uprn))

        # Click search button
        findAddress = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located(
                (By.ID, "ContentPlaceHolder1_CollectionDayLookup2_Button_SelectAddress")
            )
        )
        findAddress.click()

        h4_element = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located(
                (By.XPATH, "//h4[contains(text(), 'Collection Dates')]")
            )
        )

        soup = BeautifulSoup(driver.page_source, features="html.parser")

        bin_tables = soup.find_all("table")

        for bin_table in bin_tables:
            bin_text = bin_table.find("td", id=re.compile("CollectionDayLookup2_td_"))
            if not bin_text:
                continue

            bin_type_soup = bin_text.find("b")

            if not bin_type_soup:
                continue
            bin_type: str = bin_type_soup.text.strip().split(" (")[0]

            date_soup = bin_text.find(
                "span", id=re.compile(r"CollectionDayLookup2_Label_\w*_Date")
            )
            if not date_soup or (
                " " not in date_soup.text.strip()
                and date_soup.text.strip().lower() != "today"
            ):
                continue
            date_str: str = date_soup.text.strip()
            try:
                if date_soup.text.strip().lower() == "today":
                    date = datetime.now().date()
                else:
                    date = datetime.strptime(date_str.split(" ")[1], "%d/%m/%Y").date()

            except ValueError:
                continue

            dict_data = {
                "type": bin_type,
                "collectionDate": date.strftime("%d/%m/%Y"),
            }
            bindata["bins"].append(dict_data)

        return bindata



from contextlib import redirect_stdout
from io import StringIO
from unittest import mock
from unittest.mock import MagicMock, mock_open, patch

import pytest
from selenium.common.exceptions import WebDriverException
from uk_bin_collection.common import *
from urllib3.exceptions import MaxRetryError


def test_check_postcode_valid():
    valid_postcode = "SW1A 1AA"
    result = check_postcode(valid_postcode)
    assert result is True


def test_check_postcode_invalid():
    invalid_postcode = "BADPOSTCODE"
    with pytest.raises(ValueError) as exc_info:
        result = check_postcode(invalid_postcode)
    assert exc_info._excinfo[1].args[0] == "Exception: Invalid postcode Status: 404"
    assert exc_info.type == ValueError


def test_check_paon():
    valid_house_num = "1"
    result = check_paon(valid_house_num)
    assert result is True


def test_check_paon_invalid(capfd):
    invalid_house_num = None
    with pytest.raises(SystemExit) as exc_info:
        result = check_paon(invalid_house_num)
    out, err = capfd.readouterr()
    assert out.startswith("Exception encountered: Invalid house number")
    assert exc_info.type == SystemExit
    assert exc_info.value.code == 1


def test_get_data_check_uprn():
    uprn = "1"
    result = check_uprn(uprn)
    assert result is True


def test_get_data_check_uprn_exception(capfd):
    uprn = None
    result = check_uprn(uprn)
    out, err = capfd.readouterr()
    assert out.startswith("Exception encountered: ")


def test_get_data_check_usrn():
    usrn = "1"
    result = check_usrn(usrn)
    assert result is True


def test_get_data_check_usrn_exception(capfd):
    usrn = None
    result = check_usrn(usrn)
    out, err = capfd.readouterr()
    assert out.startswith("Exception encountered: ")


def test_get_date_with_ordinal():
    date_number = 1
    result = get_date_with_ordinal(date_number)
    assert result == "1st"


def test_get_date_with_ordinal_exception():
    date_number = "a"
    with pytest.raises(TypeError) as exc_info:
        result = get_date_with_ordinal(date_number)
    assert exc_info.type == TypeError
    assert (
        exc_info.value.args[0] == "not all arguments converted during string formatting"
    )


def test_parse_header():
    input_header = "i:am|:a:test:header|value:test"
    result = parse_header(input_header)
    assert result == {"i": "am", ":a": "test:header", "value": "test"}
    assert type(result) is dict


# Mock data for holidays
mock_holidays = {
    datetime(2023, 1, 1): "New Year's Day",
    datetime(2023, 12, 25): "Christmas Day",
    datetime(2023, 12, 26): "Boxing Day",
}


@patch("holidays.country_holidays")
def test_is_holiday_when_true(mock_holidays_func):
    # Setting up the mock to return specific holidays
    mock_holidays_func.return_value = mock_holidays

    # Christmas Day is a holiday
    assert is_holiday(datetime(2023, 12, 25), Region.ENG) is True


@patch("holidays.country_holidays")
def test_is_holiday_when_false(mock_holidays_func):
    # Setting up the mock to return specific holidays
    mock_holidays_func.return_value = mock_holidays

    # January 2nd is not a holiday
    assert is_holiday(datetime(2023, 1, 2), Region.ENG) is False


def holiday_effect(country_code, subdiv=None):
    if subdiv == "ENG":
        return {
            datetime(2023, 12, 25): "Christmas Day",
            datetime(2023, 12, 26): "Boxing Day",
        }
    elif subdiv == "SCT":
        return {datetime(2023, 11, 30): "St Andrew's Day"}
    return {}


@patch("holidays.country_holidays", side_effect=holiday_effect)
def test_is_holiday_different_region(mock_holidays_func):
    # St Andrew's Day in Scotland
    assert is_holiday(datetime(2023, 11, 30), Region.SCT) is True

    # St Andrew's Day is not observed in England
    assert is_holiday(datetime(2023, 11, 30), Region.ENG) is False


def test_is_weekend_when_true():
    weekend_date = datetime(2024, 12, 7)
    assert is_weekend(weekend_date) is True


def test_is_weekend_when_false():
    weekend_date = datetime(2024, 12, 6)
    assert is_weekend(weekend_date) is False


def test_is_working_day_when_true():
    working_day_date = datetime(2024, 12, 6)
    assert is_working_day(working_day_date) is True


def test_is_working_day_when_false():
    working_day_date = datetime(2024, 12, 7)
    assert is_working_day(working_day_date) is False


def test_get_next_working_day():
    sample_date = datetime(2024, 12, 7)
    next_working_day = get_next_working_day(sample_date)
    assert next_working_day == datetime(2024, 12, 9)


def test_remove_alpha_characters():
    test_string = "12345abc12345"
    result = remove_alpha_characters(test_string)
    assert result == "1234512345"


def test_remove_alpha_characters_bad():
    test_string = "12345abc12345"
    result = remove_alpha_characters(test_string)
    assert result != "12345abc12345"


def test_get_dates_every_x_days():
    now = datetime(2023, 2, 25, 7, 7, 17, 748661)
    result = get_dates_every_x_days(now, 5, 7)
    assert len(result) == 7
    assert result[6] == "27/03/2023"


def test_get_dates_every_x_days_bad():
    now = datetime(2023, 2, 25, 7, 7, 17, 748661)
    result = get_dates_every_x_days(now, 5, 7)
    assert len(result) != 8
    assert result[6] != "27/03/2022"


def test_remove_ordinal_indicator_from_date_string():
    test_string = "June 12th 2022"
    result = remove_ordinal_indicator_from_date_string(test_string)
    assert result == "June 12 2022"


def test_remove_ordinal_indicator_from_date_string_bad():
    test_string = "June 12th 2022"
    result = remove_ordinal_indicator_from_date_string(test_string)
    assert result != "June 12th 2022"


def test_get_weekday_dates_in_period():
    now = datetime(2023, 2, 25, 7, 7, 17, 748661)
    result = get_weekday_dates_in_period(now, 5, 7)
    assert len(result) == 7
    assert result[6] == "08/04/2023"


def test_get_weekday_dates_in_period_bad():
    now = datetime(2023, 2, 25, 7, 7, 17, 748661)
    result = get_weekday_dates_in_period(now, 5, 7)
    assert len(result) != 8
    assert result[6] != "08/04/20232"


def test_get_next_occurrence_from_day_month_false():
    result = get_next_occurrence_from_day_month(datetime(2023, 12, 1))
    assert result == datetime(2023, 12, 1, 0, 0)


def test_get_next_occurrence_from_day_month_true():
    result = get_next_occurrence_from_day_month(datetime(2023, 1, 1))
    assert result == pd.Timestamp("2024-01-01 00:00:00")


@patch("uk_bin_collection.common.load_data", return_value={})
@patch("uk_bin_collection.common.save_data")
def test_update_input_json(mock_save_data, mock_load_data):
    update_input_json(
        "test_council",
        "TEST_URL",
        "path/to/input.json",
        postcode="TEST_POSTCODE",
        uprn="TEST_UPRN",
        web_driver="TEST_WEBDRIVER",
        skip_get_url=True,
    )
    # Check that save_data was called with expected data
    expected_data = {
        "test_council": {
            "wiki_name": "test_council",
            "url": "TEST_URL",
            "postcode": "TEST_POSTCODE",
            "uprn": "TEST_UPRN",
            "web_driver": "TEST_WEBDRIVER",
            "skip_get_url": True,
        }
    }
    mock_save_data.assert_called_once_with("path/to/input.json", expected_data)


@patch("uk_bin_collection.common.load_data")
@patch("uk_bin_collection.common.save_data")
def test_update_input_json_ioerror(mock_save_data, mock_load_data):
    mock_load_data.side_effect = IOError("Unable to access file")

    with patch("builtins.print") as mock_print:
        update_input_json("test_council", "TEST_URL", "path/to/input.json")
        mock_print.assert_called_once_with(
            "Error updating the JSON file: Unable to access file"
        )


@patch("uk_bin_collection.common.load_data")
@patch("uk_bin_collection.common.save_data")
def test_update_input_json_jsondecodeerror(mock_save_data, mock_load_data):
    mock_load_data.side_effect = json.JSONDecodeError("Expecting value", "doc", 0)

    with patch("builtins.print") as mock_print:
        update_input_json("test_council", "TEST_URL", "path/to/input.json")
        mock_print.assert_called_once_with(
            "Failed to decode JSON, check the integrity of the input file."
        )


def test_load_data_existing_file():
    # Create a mock file with JSON content
    mock_file_data = json.dumps({"key": "value"})
    # Set up the mock to return a readable stream
    m = mock_open(read_data=mock_file_data)
    with patch("builtins.open", m):
        with patch("os.path.exists", return_value=True):
            data = load_data("path/to/mock/file.json")
            assert data == {
                "key": "value"
            }, f"Data was {data} instead of {{'key': 'value'}}"


def test_load_data_non_existing_file():
    # Simulate file not existing
    with patch("os.path.exists", return_value=False):
        data = load_data("path/to/nonexistent/file.json")
        assert data == {}


def test_load_data_invalid_json():
    # Create a mock file with invalid JSON content
    mock_file_data = '{"key": "value"'
    with patch("builtins.open", mock_open(read_data=mock_file_data)), patch(
        "json.load", side_effect=json.JSONDecodeError("Expecting ',' delimiter", "", 0)
    ):
        data = load_data("path/to/invalid.json")
        assert data == {}  # Modify based on your desired behavior


def test_save_data_to_file():
    # Mock the open function and simulate writing
    mock_file = mock_open()
    with patch("builtins.open", mock_file):
        data = {"key": "value"}
        save_data("path/to/mock/file.json", data)
        # Ensure the mock was called correctly to open the file for writing
        mock_file.assert_called_once_with("path/to/mock/file.json", "w")

        # Now check what was written to the file
        written_data = "".join(
            call.args[0] for call in mock_file().write.call_args_list
        )
        expected_data = json.dumps(data, sort_keys=True, indent=4)
        assert (
            written_data == expected_data
        ), "Data written to file does not match expected JSON data"


def test_save_data_io_error():
    # Simulate an IOError
    with patch("builtins.open", mock_open()) as mocked_file:
        mocked_file.side_effect = IOError("Failed to write to file")
        with pytest.raises(IOError):
            save_data("path/to/mock/file.json", {"key": "value"})


def test_contains_date_with_valid_dates():
    assert contains_date("2023-05-10")
    assert contains_date("10th of December, 2021")
    assert contains_date("March 15, 2020")
    assert contains_date("01/31/2020")


def test_contains_date_with_invalid_dates():
    assert not contains_date("not a date")
    assert not contains_date("12345")
    assert not contains_date("May 35, 2020")  # Invalid day
    assert not contains_date("2020-02-30")  # Invalid date


def test_contains_date_with_fuzzy_true():
    assert contains_date("Today is 13th of April, 2024", fuzzy=True)
    assert contains_date("They met on June 20th last year", fuzzy=True)


def test_contains_date_with_fuzzy_false():
    assert not contains_date("Today is 13th of April, 2024", fuzzy=False)
    assert not contains_date("They met on June 20th last year", fuzzy=False)


def test_contains_date_with_mixed_content():
    assert contains_date("Event starts on 2023-05-10 at 10:00 AM", fuzzy=True)
    assert not contains_date("Event starts on 2023-05-10 at 10:00 AM", fuzzy=False)


def test_create_webdriver_local():
    result = create_webdriver(
        None, headless=True, user_agent="FireFox", session_name="test-session"
    )
    assert result.name in ["chrome", "chrome-headless-shell"]


def test_create_webdriver_remote_failure():
    # Test the scenario where the remote server is not available
    with pytest.raises(MaxRetryError) as exc_info:
        create_webdriver("http://invalid-url:4444", False)


def test_create_webdriver_remote_with_session_name():
    # Test creating a remote WebDriver with a session name
    session_name = "test-session"
    web_driver_url = (
        "http://localhost:4444/wd/hub"  # Use a valid remote WebDriver URL for testing
    )

    # Mock the Remote WebDriver
    with mock.patch("uk_bin_collection.common.webdriver.Remote") as mock_remote:
        mock_instance = mock.MagicMock()
        mock_instance.name = "chrome"
        mock_remote.return_value = mock_instance

        # Call the function with the test parameters
        result = create_webdriver(web_driver=web_driver_url, session_name=session_name)

        # Check if the session name was set in capabilities
        args, kwargs = mock_remote.call_args
        options = kwargs["options"]
        assert options._caps.get("se:name") == session_name
        assert result.name == "chrome"


def test_string_with_numbers():
    assert has_numbers("abc123") is True
    assert has_numbers("1a2b3c") is True
    assert has_numbers("123") is True


def test_string_without_numbers():
    assert has_numbers("abcdef") is False
    assert has_numbers("ABC") is False
    assert has_numbers("!@#") is False


def test_empty_string():
    assert has_numbers("") is False


def test_string_with_only_numbers():
    assert has_numbers("1234567890") is True


def test_string_with_special_characters_and_numbers():
    assert has_numbers("!@#123$%^") is True
    assert has_numbers("abc!@#123") is True


def test_string_with_whitespace_and_numbers():
    assert has_numbers(" 123 ") is True
    assert has_numbers("abc 123") is True


@pytest.mark.parametrize(
    "today_str, day_name, expected",
    [
        (
            "2024-09-02",
            "Monday",
            "09/09/2024",
        ),  # Today is Monday, so next Monday is in 7 days
        (
            "2024-09-02",
            "Tuesday",
            "09/03/2024",
        ),  # Today is Monday, next Tuesday is tomorrow
        (
            "2024-09-02",
            "Sunday",
            "09/08/2024",
        ),  # Today is Monday, next Sunday is in 6 days
        (
            "2024-09-03",
            "Wednesday",
            "09/04/2024",
        ),  # Today is Tuesday, next Wednesday is tomorrow
        (
            "2024-09-03",
            "Monday",
            "09/09/2024",
        ),  # Today is Tuesday, next Monday is in 6 days
    ],
)
def test_get_next_day_of_week(today_str, day_name, expected):
    mock_today = datetime.strptime(today_str, "%Y-%m-%d")
    with patch(
        "uk_bin_collection.common.datetime"
    ) as mock_datetime:  # replace 'your_module' with the actual module name
        mock_datetime.now.return_value = mock_today
        mock_datetime.side_effect = lambda *args, **kw: datetime(*args, **kw)
        result = get_next_day_of_week(day_name, date_format="%m/%d/%Y")
        assert result == expected



import re
import time

import requests
from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_paon = kwargs.get("paon")
        user_postcode = kwargs.get("postcode")
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")
        check_paon(user_paon)
        check_postcode(user_postcode)
        bindata = {"bins": []}

        URI_1 = "https://www.hertsmere.gov.uk/Environment-Refuse-and-Recycling/Recycling--Waste/Bin-collections/Collections-and-calendar.aspx"
        URI_2 = "https://hertsmere-services.onmats.com/w/webpage/round-search"

        # Create Selenium webdriver
        driver = create_webdriver(web_driver, headless, None, __name__)
        driver.get(URI_1)

        soup = BeautifulSoup(driver.page_source, "html.parser")

        current_week = (soup.find("li", class_="current")).text.strip()

        strong = soup.find_all("strong", text=re.compile(r"^Week"))

        bin_weeks = []
        for tag in strong:
            parent = tag.parent
            bin_type = (
                (parent.text).split("-")[1].strip().replace("\xa0", " ").split(" and ")
            )
            for bin in bin_type:
                dict_data = {
                    "week": tag.text.replace("\xa0", " "),
                    "bin_type": bin,
                }
                bin_weeks.append(dict_data)

        driver.get(URI_2)

        # Wait for the postcode field to appear then populate it
        inputElement_postcode = WebDriverWait(driver, 30).until(
            EC.presence_of_element_located(
                (
                    By.CLASS_NAME,
                    "relation_path_type_ahead_search",
                )
            )
        )
        inputElement_postcode.send_keys(user_postcode)

        WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.XPATH,
                    f"//ul[@class='result_list']/li[starts-with(@aria-label, '{user_paon}')]",
                )
            )
        ).click()

        WebDriverWait(driver, timeout=10).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    "input.fragment_presenter_template_edit.btn.bg-primary.btn-medium[type='submit']",
                )
            )
        ).click()

        WebDriverWait(driver, timeout=10).until(
            EC.presence_of_element_located(
                (By.XPATH, "//h3[contains(text(), 'Collection days')]")
            )
        )

        soup = BeautifulSoup(driver.page_source, "html.parser")

        table = soup.find("table", class_="table listing table-striped")

        # Check if the table was found
        if table:
            # Extract table rows and cells
            table_data = []
            for row in table.find("tbody").find_all("tr"):
                # Extract cell data from each <td> tag
                row_data = [cell.get_text(strip=True) for cell in row.find_all("td")]
                table_data.append(row_data)

        else:
            print("Table not found.")

        collection_day = (table_data[0])[1]

        current_week_bins = [bin for bin in bin_weeks if bin["week"] == current_week]
        next_week_bins = [bin for bin in bin_weeks if bin["week"] != current_week]

        days_of_week = [
            "Monday",
            "Tuesday",
            "Wednesday",
            "Thursday",
            "Friday",
            "Saturday",
            "Sunday",
        ]

        today = datetime.now()
        today_idx = today.weekday()  # Monday is 0 and Sunday is 6
        target_idx = days_of_week.index(collection_day)

        days_until_target = (target_idx - today_idx) % 7
        if days_until_target == 0:
            next_day = today
        else:
            next_day = today + timedelta(days=days_until_target)

        current_week_dates = get_dates_every_x_days(next_day, 14, 7)
        next_week_date = next_day + timedelta(days=7)
        next_week_dates = get_dates_every_x_days(next_week_date, 14, 7)

        for date in current_week_dates:
            for bin in current_week_bins:
                dict_data = {
                    "type": bin["bin_type"],
                    "collectionDate": date,
                }
                bindata["bins"].append(dict_data)

        for date in next_week_dates:
            for bin in next_week_bins:
                dict_data = {
                    "type": bin["bin_type"],
                    "collectionDate": date,
                }
                bindata["bins"].append(dict_data)

        bindata["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), "%d/%m/%Y")
        )

        return bindata



from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Get and check UPRN
        user_postcode = kwargs.get("postcode")
        user_paon = kwargs.get("paon")
        check_paon(user_paon)
        check_postcode(user_postcode)
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")
        bindata = {"bins": []}

        API_URL = "https://uhte-wrp.whitespacews.com"

        # Create Selenium webdriver
        driver = create_webdriver(web_driver, headless, None, __name__)
        driver.get(API_URL)

        # Click Find my bin collection day button
        collectionButton = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable((By.LINK_TEXT, "Find my bin collection day"))
        )
        collectionButton.click()

        main_content = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "main-content"))
        )

        # Wait for the property number field to appear then populate it
        inputElement_number = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.ID,
                    "address_name_number",
                )
            )
        )
        inputElement_number.send_keys(user_paon)

        # Wait for the postcode field to appear then populate it
        inputElement_postcode = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.ID,
                    "address_postcode",
                )
            )
        )
        inputElement_postcode.send_keys(user_postcode)

        # Click search button
        continueButton = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.ID,
                    "Submit",
                )
            )
        )
        continueButton.click()

        # Wait for the 'Search Results' to appear and select the first result
        property = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.CSS_SELECTOR,
                    "li.app-subnav__section-item a",
                    # "app-subnav__link govuk-link clicker colordarkblue fontfamilyArial fontsize12rem",
                    # "//a[starts-with(@aria-label, '{user_paon}')]",
                )
            )
        )
        property.click()

        upcoming_scheduled_collections = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "upcoming-scheduled-collections"))
        )

        soup = BeautifulSoup(driver.page_source, features="html.parser")

        collections = []
        for collection in soup.find_all(
            "u1",
            class_="displayinlineblock justifycontentleft alignitemscenter margin0 padding0",
        ):
            date = collection.find(
                "p", string=lambda text: text and "/" in text
            ).text.strip()  # Extract date
            service = collection.find(
                "p", string=lambda text: text and "Collection Service" in text
            ).text.strip()  # Extract service type
            collections.append({"date": date, "service": service})

        # Print the parsed data
        for item in collections:

            dict_data = {
                "type": item["service"],
                "collectionDate": item["date"],
            }
            bindata["bins"].append(dict_data)

        return bindata



import time
from datetime import datetime

from bs4 import BeautifulSoup
from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Get and check UPRN
        user_postcode = kwargs.get("postcode")
        user_paon = kwargs.get("paon")
        check_paon(user_paon)
        check_postcode(user_postcode)
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")
        bindata = {"bins": []}

        API_URL = "https://portal.digital.ashfield.gov.uk/w/webpage/raise-case?service=bin_calendar"

        # Create Selenium webdriver
        driver = create_webdriver(web_driver, headless, None, __name__)
        driver.get(API_URL)

        title = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "sub_page_title"))
        )

        # Wait for the postcode field to appear then populate it
        WebDriverWait(driver, 10).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "input.relation_path_type_ahead_search")
            )
        )

        inputElement_postcode = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located(
                (By.CSS_SELECTOR, "input.relation_path_type_ahead_search")
            )
        )
        inputElement_postcode.clear()
        inputElement_postcode.send_keys(user_postcode)

        # Wait for the 'Select your property' dropdown to appear and select the first result
        dropdown = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.CLASS_NAME,
                    "result_list ",
                )
            )
        )

        address_element = (
            WebDriverWait(driver, 10)
            .until(
                EC.element_to_be_clickable(
                    (By.XPATH, f"//li[starts-with(@aria-label, '{user_paon}')]")
                )
            )
            .click()
        )

        search_button = WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (By.XPATH, "//input[@type='submit' and @value='Search']")
            )
        )
        search_button.click()

        time.sleep(10)

        soup = BeautifulSoup(driver.page_source, features="html.parser")
        soup.prettify()

        # Find the table by class name
        table = soup.find("table", {"class": "table listing table-striped"})

        # Iterate over each row in the tbody of the table
        for row in table.find("tbody").find_all("tr"):
            # Extract the service, day, and date for each row
            service = row.find_all("td")[0].get_text(strip=True)
            date = row.find_all("td")[2].get_text(strip=True)

            dict_data = {
                "type": service,
                "collectionDate": datetime.strptime(date, "%a, %d %b %Y").strftime(
                    date_format
                ),
            }
            bindata["bins"].append(dict_data)

        return bindata



from urllib.parse import quote, urljoin

from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    BASE_URL = "https://www.midlothian.gov.uk"
    DIRECTORY_URL = f"{BASE_URL}/site/scripts/directory_search.php?directoryID=35&keywords={{}}&search=Search"
    BIN_TYPES = {
        "Next recycling collection": "Recycling",
        "Next grey bin collection": "Grey Bin",
        "Next brown bin collection": "Brown Bin",
        "Next food bin collection": "Food Bin",
    }

    def parse_data(self, page: str, **kwargs) -> dict:

        house_identifier = kwargs.get(
            "paon", ""
        ).strip()  # Could be house number or name
        postcode = kwargs.get("postcode")

        # Check if both house identifier and postcode are provided
        if not house_identifier:
            print("Error: House identifier (number or name) must be provided.")
            return {"bins": []}

        if not postcode:
            print("Error: Postcode must be provided.")
            return {"bins": []}

        check_postcode(postcode)
        check_paon(house_identifier)

        data = {"bins": []}
        search_url = self.DIRECTORY_URL.format(quote(postcode))

        try:
            search_results_html = requests.get(search_url)
            search_results_html.raise_for_status()

            soup = BeautifulSoup(search_results_html.text, "html.parser")
            address_link = self._get_result_by_identifier(soup, house_identifier)

            if address_link:
                collections_url = urljoin(search_url, address_link["href"])
                bin_collection_data = self._fetch_bin_collection_data(collections_url)

                if bin_collection_data:
                    data["bins"].extend(bin_collection_data)

        except requests.RequestException as e:
            print(f"Warning: Failed to fetch data from {search_url}. Error: {e}")

        return data

    def _get_result_by_identifier(self, soup, identifier: str) -> list:
        """Extract the result link that matches the given house number or house name."""
        try:
            results_list = (
                soup.find("article", class_="container")
                .find("h2", text="Search results")
                .find_next("ul", class_="item-list item-list__rich")
            )

            pattern = re.compile(re.escape(identifier.lower()) + r"[ ,]")

            for item in results_list.find_all("li"):
                address_link = item.find("a")
                if address_link:
                    link_text = address_link.text.strip().lower()
                    if pattern.match(link_text):
                        return address_link

            print(f"Warning: No results found for identifier '{identifier}'.")
            return None  # Return None if no match is found

        except AttributeError as e:
            print(f"Warning: Could not find the search results. Error: {e}")
            return None  # Return None if no result found

    def _fetch_bin_collection_data(self, url: str) -> list:
        """Fetch and parse bin collection data from the given URL."""
        try:
            bin_collection_html = requests.get(url)
            bin_collection_html.raise_for_status()

            soup = BeautifulSoup(bin_collection_html.text, "html.parser")
            bin_collections = soup.find("ul", class_="data-table")

            if bin_collections:
                return self._parse_bin_collection_items(
                    bin_collections.find_all("li")[2:]  # Skip the first two items
                )

        except requests.RequestException as e:
            print(
                f"Warning: Failed to fetch bin collection data from {url}. Error: {e}"
            )

        return []  # Return an empty list on error

    def _parse_bin_collection_items(self, bin_items: list) -> list:
        """Parse bin collection items into a structured format."""
        parsed_bins = []

        for bin_item in bin_items:
            bin_type = None
            try:
                if bin_item.h2 and bin_item.h2.text.strip() in self.BIN_TYPES:
                    bin_type = self.BIN_TYPES[bin_item.h2.text.strip()]

                bin_collection_date = None
                if bin_item.div and bin_item.div.text.strip():
                    try:
                        bin_collection_date = datetime.strptime(
                            bin_item.div.text.strip(), "%A %d/%m/%Y"
                        ).strftime(date_format)
                    except ValueError:
                        print(
                            f"Warning: Date parsing failed for {bin_item.div.text.strip()}."
                        )

                if bin_type and bin_collection_date:
                    parsed_bins.append(
                        {
                            "type": bin_type,
                            "collectionDate": bin_collection_date,
                        }
                    )
                else:
                    print(f"Warning: Missing data for bin item: {bin_item}")

            except Exception as e:
                print(f"Warning: An error occurred while parsing bin item. Error: {e}")

        return parsed_bins



import time

import requests
from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        API_URL = "https://maps.norwich.gov.uk/arcgis/rest/services/MyNorwich/PropertyDetails/FeatureServer/2/query"

        params = {
            "f": "json",
            "where": f"UPRN='{user_uprn}' or UPRN='0{user_uprn}'",
            "returnGeometry": "true",
            "spatialRel": "esriSpatialRelIntersects",
            "geometryType": "esriGeometryPolygon",
            "inSR": "4326",
            "outFields": "*",
            "outSR": "4326",
            "resultRecordCount": "1000",
        }

        r = requests.get(API_URL, params=params)

        data = r.json()
        data = data["features"][0]["attributes"]["WasteCollectionHtml"]
        soup = BeautifulSoup(data, "html.parser")

        alternateCheck = soup.find("p")
        if alternateCheck.text.__contains__("alternate"):
            alternateCheck = True
        else:
            alternateCheck = False

        strong = soup.find_all("strong")
        collections = []

        if alternateCheck:
            bin_types = strong[2].text.strip().replace(".", "").split(" and ")
            for bin in bin_types:
                collections.append((bin.capitalize(), datetime.strptime(strong[1].text.strip(), date_format)))

        else:
            p_tag = soup.find_all("p")
            i = 1
            for p in p_tag:
                bin_types = (
                    p.text.split("Your ")[1].split(" is collected")[0].split(" and ")
                )
                for bin in bin_types:
                    collections.append((bin.capitalize(), datetime.strptime(strong[1].text.strip(), date_format)))
                i += 2

        if len(strong) > 3:
            collections.append(("Garden", datetime.strptime(strong[4].text.strip(), date_format)))

        ordered_data = sorted(collections, key=lambda x: x[1])
        for item in ordered_data:
            dict_data = {
                "type": item[0] + " bin",
                "collectionDate": item[1].strftime(date_format),
            }
            bindata["bins"].append(dict_data)

        return bindata



import requests
from bs4 import BeautifulSoup

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:

        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)
        bindata = {"bins": []}

        URI = f"https://www.wandsworth.gov.uk/my-property/?UPRN={user_uprn}"

        # Make the GET request
        response = requests.get(URI)

        soup = BeautifulSoup(response.content, features="html.parser")
        soup.prettify()

        # Find all collection types
        collection_types = soup.find_all("h4", class_="collection-heading")

        # Iterate over each collection type
        for collection_type in collection_types:
            bin_types = collection_type.text.strip().split("/")
            collections = collection_type.find_next_sibling("div", class_="collections")

            # Extract next and previous collections
            next_collection = collections.find_all("div", class_="collection")

            # Parse each collection
            for collection in next_collection:
                # Extract the collection type (Next or Previous)
                strong_tag = collection.find("strong")
                collection_type = (
                    strong_tag.text.strip(":") if strong_tag else "Unknown"
                )

                # Extract the date
                date_text = (
                    strong_tag.next_sibling.strip()
                    if strong_tag and strong_tag.next_sibling
                    else "No date found"
                )

                if date_text == "No date found":
                    continue

                for bin_type in bin_types:
                    # Append to the schedule
                    dict_data = {
                        "type": bin_type,
                        "collectionDate": datetime.strptime(
                            date_text,
                            "%A %d %B %Y",
                        ).strftime(date_format),
                    }
                    bindata["bins"].append(dict_data)

        bindata["bins"].sort(
            key=lambda x: datetime.strptime(x.get("collectionDate"), date_format)
        )

        return bindata



import json

import requests
from bs4 import BeautifulSoup
from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


# import the wonderful Beautiful Soup and the URL grabber
class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        user_uprn = kwargs.get("uprn")
        check_uprn(user_uprn)

        headers = {
            "authority": "www.wealden.gov.uk",
            "accept": "*/*",
            "accept-language": "en-GB,en;q=0.7",
            "content-type": "application/x-www-form-urlencoded; charset=UTF-8",
            # Requests sorts cookies= alphabetically
            # 'cookie': 'ARRAffinity=e45c20b343b490e3866d5d35c3dbda687e4a1357c2163c32922209862abb5872; ARRAffinitySameSite=e45c20b343b490e3866d5d35c3dbda687e4a1357c2163c32922209862abb5872',
            "origin": "https://www.wealden.gov.uk",
            "referer": "https://www.wealden.gov.uk/recycling-and-waste/bin-search/?uprn=10033413624",
            "sec-fetch-dest": "empty",
            "sec-fetch-mode": "cors",
            "sec-fetch-site": "same-origin",
            "sec-gpc": "1",
            "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 Safari/537.36",
            "x-requested-with": "XMLHttpRequest",
        }

        data = {
            "action": "wealden_get_collections_for_uprn",
            "uprn": user_uprn,
        }

        requests.packages.urllib3.disable_warnings()
        response = requests.post(
            "https://www.wealden.gov.uk/wp-admin/admin-ajax.php",
            headers=headers,
            data=data,
        )
        json_data = json.loads(response.text)

        if json_data["status"] != "success":
            raise ValueError("Error parsing data. Please open an issue on GitHub.")

        property_data = json_data["collection"]
        data = {"bins": []}
        collections = []

        if len(property_data["refuseCollectionDate"]) > 0:
            bin_type = "Rubbish Bin"
            bin_date = datetime.strptime(
                property_data["refuseCollectionDate"], "%Y-%m-%dT%H:%M:%S"
            )
            collections.append((bin_type, bin_date))
        if len(property_data["recyclingCollectionDate"]) > 0:
            bin_type = "Recycling Bin"
            bin_date = datetime.strptime(
                property_data["recyclingCollectionDate"], "%Y-%m-%dT%H:%M:%S"
            )
            collections.append((bin_type, bin_date))
        if len(property_data["gardenCollectionDate"]) > 0:
            bin_type = "Garden Bin"
            bin_date = datetime.strptime(
                property_data["gardenCollectionDate"], "%Y-%m-%dT%H:%M:%S"
            )
            collections.append((bin_type, bin_date))

        ordered_data = sorted(collections, key=lambda x: x[1])
        data = {"bins": []}
        for item in ordered_data:
            dict_data = {
                "type": item[0],
                "collectionDate": item[1].strftime(date_format),
            }
            data["bins"].append(dict_data)

        return data



from bs4 import BeautifulSoup
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

from uk_bin_collection.uk_bin_collection.common import *
from uk_bin_collection.uk_bin_collection.get_bin_data import AbstractGetBinDataClass


def parse_collection_date(date_string) -> datetime:
    now = datetime.now()
    if date_string == "is due today":
        return now

    parsed_date = datetime.strptime(date_string, "%A, %d %B").replace(year=now.year)

    if now.month == 12 and parsed_date.month < 12:
        parsed_date = parsed_date.replace(year=(now.year + 1))

    return parsed_date

class CouncilClass(AbstractGetBinDataClass):
    """
    Concrete classes have to implement all abstract operations of the
    base class. They can also override some operations with a default
    implementation.
    """

    def parse_data(self, page: str, **kwargs) -> dict:
        # Get postcode and UPRN from kwargs
        user_postcode = kwargs.get("postcode")
        user_paon = kwargs.get("paon")
        web_driver = kwargs.get("web_driver")
        headless = kwargs.get("headless")
        check_postcode(user_postcode)
        check_paon(user_paon)

        # Build URL to parse
        council_url = "https://swale.gov.uk/bins-littering-and-the-environment/bins/check-your-bin-day"

        # Create Selenium webdriver
        driver = create_webdriver(web_driver, headless, None, __name__)
        driver.get(council_url)

        # Wait for the postcode field to appear then populate it
        try:
            inputElement_postcode = WebDriverWait(driver, 10).until(
                EC.presence_of_element_located((By.ID, "q485476_q1"))
            )
            inputElement_postcode.send_keys(user_postcode)
        except Exception:
            print("Page failed to load. Probably due to Cloudflare robot check!")

        # Click search button
        findAddress = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "form_email_485465_submit"))
        )
        driver.execute_script("arguments[0].click();", findAddress)

        # Wait for the 'Select address' dropdown to appear and select option matching the house name/number
        WebDriverWait(driver, 10).until(
            EC.element_to_be_clickable(
                (
                    By.XPATH,
                    "//select[@name='q485480:q1']//option[contains(., '"
                    + user_paon
                    + "')]",
                )
            )
        ).click()

        # Click search button
        getBins = WebDriverWait(driver, 10).until(
            EC.presence_of_element_located((By.ID, "form_email_485465_submit"))
        )
        driver.execute_script("arguments[0].click();", getBins)

        BinTable = WebDriverWait(driver, 30).until(
            EC.presence_of_element_located((By.ID, "SBCYBDSummary"))
        )

        soup = BeautifulSoup(driver.page_source, features="html.parser")
        soup.prettify()

        data = {"bins": []}

        next_collection_date = soup.find(
            "strong", id="SBC-YBD-collectionDate"
        ).text.strip()

        # Extract bins for the next collection
        next_bins = [li.text.strip() for li in soup.select("#SBCFirstBins ul li")]

        # Extract future collection details
        future_collection_date_tag = soup.find(
            "p", text=lambda t: t and "starting from" in t
        )
        future_collection_date = (
            future_collection_date_tag.text.split("starting from")[-1].strip()
            if future_collection_date_tag
            else "No future date found"
        )

        future_bins = [li.text.strip() for li in soup.select("#FirstFutureBins li")]

        for bin in next_bins:
            dict_data = {
                "type": bin,
                "collectionDate": datetime.strptime(
                    next_collection_date, "%A, %d %B"
                ).strftime(date_format),
            }
            data["bins"].append(dict_data)

        for bin in future_bins:
            dict_data = {
                "type": bin,
                "collectionDate": datetime.strptime(
                    future_collection_date, "%A, %d %B"
                ).strftime(date_format),
            }
            data["bins"].append(dict_data)

        return data
